import pydantic


class EngineerException(Exception):
    """
    The base class for errors in :mod:`royalnet.engineer`.
    """


class WrenchException(EngineerException):
    """
    The base class for errors in :mod:`royalnet.engineer.wrench`.
    """


class DeliberateException(WrenchException):
    """
    This exception was deliberately raised by :class:`royalnet.engineer.wrench.ErrorAll`.
    """


class TeleporterError(EngineerException, pydantic.ValidationError):
    """
    The base class for errors in :mod:`royalnet.engineer.teleporter`.
    """


class InTeleporterError(TeleporterError):
    """
    The input parameters validation failed.
    """


class OutTeleporterError(TeleporterError):
    """
    The return value validation failed.
    """


class BulletException(EngineerException):
    """
    The base class for errors in :mod:`royalnet.engineer.bullet`.
    """


class NotSupportedError(BulletException, NotImplementedError):
    """
    The requested property isn't available on the current frontend.
    """


__all__ = (
    "EngineerException",
    "WrenchException",
    "DeliberateException",
    "TeleporterError",
    "InTeleporterError",
    "OutTeleporterError",
    "BulletException",
    "NotSupportedError",
)



import pytest
from ..asynccampaign import AsyncCampaign
from ..asyncchallenge import AsyncChallenge
from ..exc import *


@pytest.mark.asyncio
async def test_creation():
    async def gen():
        yield

    await AsyncCampaign.create(start=gen())


@pytest.mark.asyncio
async def test_send_receive():
    async def gen():
        ping = yield
        assert ping == "Ping!"
        yield None, "Pong!"

    campaign = await AsyncCampaign.create(start=gen())
    pong, = await campaign.next("Ping!")
    assert pong == "Pong!"


class FalseChallenge(AsyncChallenge):
    async def filter(self, data) -> bool:
        return False


@pytest.mark.asyncio
async def test_failing_check():
    async def gen():
        yield

    campaign = await AsyncCampaign.create(start=gen(), challenge=FalseChallenge())
    with pytest.raises(ChallengeFailedError):
        await campaign.next()


@pytest.mark.asyncio
async def test_switching():
    async def gen_1():
        yield
        yield gen_2()

    async def gen_2():
        yield
        yield None, "Second message!"
        yield None

    campaign = await AsyncCampaign.create(start=gen_1())
    data, = await campaign.next()
    assert data == "Second message!"



from __future__ import annotations
from royalnet.royaltyping import *
import logging
import inspect
import datetime
from .asyncchallenge import AsyncChallenge, TrueAsyncChallenge
from .exc import *
log = logging.getLogger(__name__)

__all__ = (
    "AsyncCampaign",
)


class AsyncCampaign:
    """
    The AsyncCampaign module allows for branching asyncgenerator-based back-and-forths between the software and the
    user.

    An :class:`.AsyncCampaign` consists of multiple chained AsyncAdventures, which are AsyncGenerators yielding tuples
    with an AsyncChallenge and optional data.
    """
    def __init__(self, start: AsyncAdventure, challenge: Optional[AsyncChallenge], *args, **kwargs):
        """
        Initialize an :class:`.AsyncCampaign` object.

        .. warning:: Do not use this, use the :meth:`.create()` method instead!

        :param start: The starting adventure for the :class:`.AsyncCampaign`.
        """
        self.adventure: AsyncAdventure = start
        self.challenge: AsyncChallenge = challenge or TrueAsyncChallenge()
        self.last_update: datetime.datetime = ...

    @classmethod
    async def create(cls, start: AsyncAdventure, challenge: Optional[AsyncChallenge] = None, *args, **kwargs) -> AsyncCampaign:
        """
        Create a new :class:`.AsyncCampaign` object.

        :param start: The starting Adventure for the :class:`.AsyncCampaign`.
        :param challenge: The AsyncChallenge the campaign should start with.
        :return: The created :class:`.AsyncCampaign`.
        """
        campaign = cls(start=start, challenge=challenge, *args, **kwargs)
        await campaign._asend(None)
        return campaign

    async def _asend(self, data: Any) -> Any:
        try:
            return await self.adventure.asend(data)
        except RuntimeError:
            log.error(f"{self.adventure} is being used unexpectedly by something else!")
            raise

    async def _athrow(self, typ: Type[BaseException], val: Optional[BaseException], tb: Any) -> Any:
        try:
            return await self.adventure.athrow(typ, val, tb)
        except RuntimeError:
            log.error(f"{self.adventure} is being used unexpectedly by something else!")
            raise

    async def _aclose(self) -> None:
        try:
            await self.adventure.aclose()
        except RuntimeError:
            log.error(f"{self.adventure} is being used unexpectedly by something else!")
            raise

    async def next(self, data: Any = None) -> List:
        """
        Try to advance the :class:`.AsyncCampaign` with the passed data.

        :param data: The data to pass to the current AsyncAdventure.
        :return: Optional additional data returned by the AsyncAdventure.
        :raises ChallengeFailedError: if the data passed fails the AsyncChallenge check.
        """
        self.last_update = datetime.datetime.now()
        if not await self.challenge.filter(data):
            raise ChallengeFailedError(f"{data} failed the {self.challenge} challenge")
        result = await self._asend(data)
        if inspect.isasyncgen(result):
            await self._aclose()
            self.adventure = result
            await self._asend(None)
            return await self.next(data)
        elif isinstance(result, AsyncChallenge):
            self.challenge = result
            return []
        elif result is None:
            return []
        elif isinstance(result, tuple) and len(result) > 0:
            if isinstance(result[0], AsyncChallenge):
                self.challenge, *output = result
                return output
            elif result[0] is None:
                _, *output = result
                return output
        else:
            raise TypeError(f"AsyncAdventure yielded an invalid type: {result.__class_}")



"""
This module defines multiple classes that can be used to facilitate the implementation of a conversational interface
in chat bots.
"""

from .campaign import *
from .asynccampaign import *
from .challenge import *
from .asyncchallenge import *


__all__ = (
    "Campaign",
    "AsyncCampaign",
    "Challenge",
    "TrueChallenge",
    "AsyncChallenge",
    "TrueAsyncChallenge",
)



"""
Commands are used to quickly create single-message conversations
"""

from __future__ import annotations
import royalnet.royaltyping as t

import logging
import re

from . import teleporter

log = logging.getLogger(__name__)


class Command:
    """
    A class that allows creating commands which can be called from the chat by entering a certain :attr:`.pattern` of
    characters:

    .. code-block:: text

        /echo Hello!
        # Hello!

    .. todo:: Improve the docstring of :class:`.Command`.
    """

    def __init__(self,
                 prefix: str,
                 name: str,
                 syntax: str,
                 conversation: t.Conversation,
                 *,
                 pattern: str = r"^{prefix}{name} ?{syntax}",
                 doc: str = ""):
        self.prefix: str = prefix
        """
        The prefix used in the command (usually ``/`` or ``!``).
        """

        self.name: str = name
        """
        The name of the command, usually all lowercase.
        """

        self.syntax: str = syntax
        """
        A regex describing the syntax of the command, using named capture groups ``(?P<name>...)`` to capture arguments
        that should be passed to the function.
        """

        self.pattern: re.Pattern = re.compile(pattern.format(prefix=prefix, name=name, syntax=syntax))
        """
        The compiled regex pattern. 
        
        By default, it :meth:`str.format` the passed string with the ``prefix``, ``name`` and ``syntax`` keyword 
        arguments, but this behaviour can be changed by passing a different ``pattern`` to :meth:`__init__`.
        """

        self.doc: str = doc
        """
        A string explaining how this command should be used. Useful for command lists or help commands.
        """

        self.conversation: t.Conversation = conversation
        """
        The conversation to run when this command is called.
        """

    @classmethod
    def new(cls,
            prefix: str,
            name: str,
            syntax: str,
            *,
            pattern: str = r"^{prefix}{name} {syntax}",
            doc: str = ""):
        """
        Create a new :class:`.Command` using the decorated function as :attr:`.conversation`.
        :return: The created :class:`.Command`.
        """
        def decorator(f):
            log.debug(f"Making function {f} a teleporter...")
            teleporter_f = teleporter.teleport(is_async=True, validate_output=False)(f)

            log.debug(f"Creating command: {prefix} {name} {syntax} - {doc}")
            return cls(prefix=prefix, name=name, syntax=syntax, conversation=teleporter_f, pattern=pattern, doc=doc)
        return decorator

    async def run(self, _text: str, **original_kwargs) -> t.Optional[t.Conversation]:
        """
        Run the command.

        :param _text: The text of the message that was received.
        """
        log.debug(f"Matching text {_text} to {self.pattern}...")
        match: re.Match = self.pattern.search(_text)
        if match is None:
            log.debug(f"Pattern didn't match, returning...")
            return

        log.debug(f"Pattern matched, getting named groups...")
        match_kwargs: dict = match.groupdict()

        log.debug(f"Running teleported function with args: {match_kwargs}")
        return await self.conversation(**original_kwargs, **match_kwargs)


__all__ = (
    "Command",
)



from royalnet.royaltyping import *
import abc

__all__ = (
    "AsyncChallenge",
    "TrueAsyncChallenge",
)


class AsyncChallenge(metaclass=abc.ABCMeta):
    """A filter for inputs passed to an :class:`.AsyncCampaign`."""

    @abc.abstractmethod
    async def filter(self, data: Any) -> bool:
        """Decide if the data should be skipped or not."""
        raise NotImplementedError()


class TrueAsyncChallenge(AsyncChallenge):
    """An :class:`.AsyncChallenge` which always returns :data:`True`."""

    async def filter(self, data: Any) -> bool:
        """Decide if the data should be skipped or not."""
        return True



from __future__ import annotations
from royalnet.royaltyping import *
import logging
import inspect
import datetime
from .challenge import Challenge, TrueChallenge
from .exc import *
log = logging.getLogger(__name__)

__all__ = (
    "Campaign",
)


class Campaign:
    """
    The Campaign module allows for branching generator-based back-and-forths between the software and the
    user.

    A :class:`.Campaign` consists of multiple chained Adventures, which are Generators yielding tuples with a
    :class:`.Campaign` and optional data.
    """

    def __init__(self, start: Adventure, challenge: Optional[Challenge] = None, *args, **kwargs):
        """
        Initialize a :class:`.Campaign` object.

        .. warning:: Do not use this, use the :meth:`.create` method instead!
        """
        self.adventure: Adventure = start
        self.challenge: Challenge = challenge or TrueChallenge()
        self.last_update: datetime.datetime = ...

    @classmethod
    def create(cls, start: Adventure, challenge: Optional[Challenge] = None, *args, **kwargs) -> Campaign:
        """
        Create a new :class:`.Campaign` object.

        :param start: The starting Adventure for the :class:`.Campaign`.
        :param challenge: The Challenge the :class:`.Campaign` should start with.
        :return: The created :class:`.Campaign`.
        """
        campaign = cls(start=start, challenge=challenge, *args, **kwargs)
        campaign.adventure.send(None)
        return campaign

    def next(self, data: Any = None) -> List:
        """
        Try to advance the :class:`.Campaign` with the passed data.

        :param data: The data to pass to the current Adventure.
        :return: Optional additional data returned by the Adventure.
        :raises .ChallengeFailedError: if the data passed fails the Challenge check.
        """
        self.last_update = datetime.datetime.now()
        if not self.challenge.filter(data):
            raise ChallengeFailedError(f"{data} failed the {self.challenge} challenge")
        result = self.adventure.send(data)
        if inspect.isgenerator(result):
            self.adventure.close()
            self.adventure = result
            self.adventure.send(None)
            return self.next(data)
        elif isinstance(result, Challenge):
            self.challenge = result
            return []
        elif result is None:
            return []
        elif isinstance(result, tuple) and len(result) > 0:
            if isinstance(result[0], Challenge):
                self.challenge, *output = result
                return output
            elif result[0] is None:
                _, *output = result
                return output
        else:
            raise TypeError(f"Adventure yielded an invalid type: {result.__class_}")



"""
This module defines a class that can be used to easily load configuration from a :mod:`json` or :mod:`toml` file and
from :data:`os.environ`.
"""

from __future__ import annotations
from royalnet.royaltyping import *
import os
import re
import toml
import json

from .exc import *


T = TypeVar("T")


class Scroll:
    """
    A configuration handler that allows getting values from both the environment variables and a config file.
    """

    key_validator = re.compile(r"^[A-Za-z.]+$")
    """
    A regex used to validate config keys.
    """

    def __init__(self, namespace: str, config: Optional[Dict[str, JSON]] = None):
        self.namespace: str = namespace
        self.config: Optional[Dict[str, JSON]] = config

    @classmethod
    def from_toml(cls, namespace: str, file_path: os.PathLike) -> Scroll:
        """
        Create a new :class:`.Scroll` from a :mod:`toml` file.

        :param namespace: A string prefixed to the environment variable keys to disambiguate between multiple scroll
                          objects.
        :param file_path: The path of the :mod:`toml` file to load.
        :return: The created :class:`.Scroll` object.
        """
        with open(file_path) as file:
            config = toml.load(file)
        return cls(namespace, config)

    @classmethod
    def from_json(cls, namespace: str, file_path: os.PathLike) -> Scroll:
        """
        Create a new :class:`.Scroll` from a :mod:`json` file.

        :param namespace: A string prefixed to the environment variable keys to disambiguate between multiple scroll
                          objects.
        :param file_path: The path of the :mod:`json` file to load.
        :return: The created :class:`.Scroll` object.
        """
        with open(file_path) as file:
            config = json.load(file)
        return cls(namespace, config)

    @classmethod
    def from_file(cls, namespace: str, file_path: os.PathLike, require_file: bool = False) -> Scroll:
        """
        Try to guess the type of the config file and create a new :class:`.Scroll` from it.

        :param namespace: A string prefixed to the environment variable keys to disambiguate between multiple scroll
                          objects.
        :param file_path: The path of the file to load.
        :param require_file: Require the file to exist for the :class:`.Scroll` to be created.
        :raise .exc.InvalidFileType: If the file format isn't supported.
        :raise .exc.MissingConfigFile: If ``require_file`` is set to :data:`True` and the config file does not exist.
        :return: The created :class:`.Scroll` object.
        """
        if require_file and not os.path.exists(file_path):
            raise MissingConfigFileError(f"No config file exists at path {file_path}, and require_file is set to True.")

        _, ext = os.path.splitext(file_path)
        lext = ext.lower()

        try:
            return cls.loaders[lext](namespace=namespace, file_path=file_path)
        except KeyError:
            raise InvalidFileType(f"Invalid extension: {lext}")
        except FileNotFoundError:
            return cls(namespace=namespace)

    @classmethod
    def _validate_key(cls, item: str):
        check = cls.key_validator.match(item)
        if not check:
            raise InvalidFormatError()

    def _get_from_environ(self, item: str) -> JSONScalar:
        """Get a configuration value from the environment variables."""
        key = f"{self.namespace}_{item.replace('.', '_')}".upper()

        try:
            j = os.environ[key]
        except KeyError:
            raise NotFoundError(f"'{key}' was not found in the environment variables.")

        try:
            value = json.loads(j)
        except json.JSONDecodeError:
            raise ParseError(f"'{key}' contains invalid JSON: {j}")

        return value

    def _get_from_config(self, item: str) -> JSONScalar:
        """Get a configuration value from the configuration file."""
        if self.config is None:
            raise NotFoundError("No config file has been loaded.")

        chain = item.lower().split(".")

        current = self.config

        for element in chain:
            try:
                current = current[element]
            except KeyError:
                raise NotFoundError(f"'{item}' was unreachable: could not find '{element}'")

        return current

    def __getitem__(self, item: str):
        self._validate_key(item)
        try:
            return self._get_from_environ(item)
        except NotFoundError:
            return self._get_from_config(item)

    def get(self, item: str, default: T) -> Union[JSONScalar, T]:
        try:
            return self[item]
        except NotFoundError:
            return default


Scroll.loaders = {
    ".json": Scroll.from_json,
    ".toml": Scroll.from_toml,
}
"""
An extension to deserialization function map.
"""


__all__ = ("Scroll",)



"""
The teleporter uses :mod:`pydantic` to validate function parameters and return values.
"""

from __future__ import annotations
import royalnet.royaltyping as t

import logging
import pydantic
import inspect
import functools

from . import exc

Value = t.TypeVar("Value")
log = logging.getLogger(__name__)


class TeleporterConfig(pydantic.BaseConfig):
    """
    A :mod:`pydantic` model config which allows for arbitrary types.
    """
    arbitrary_types_allowed = True


def parameter_to_field(param: inspect.Parameter, **kwargs) -> t.Tuple[type, pydantic.fields.FieldInfo]:
    """
    Convert a :class:`inspect.Parameter` to a type-field :class:`tuple`, which can be easily passed to
    :func:`pydantic.create_model`.

    If the parameter is already a :class:`pydantic.FieldInfo` (created by :func:`pydantic.Field`), it will be
    returned as the value, without creating a new model.

    :param param: The :class:`inspect.Parameter` to convert.
    :param kwargs: Additional kwargs to pass to the field.
    :return: A :class:`tuple`, where the first element is a :class:`type` and the second is a :class:`pydantic.Field`.
    """
    if isinstance(param.default, pydantic.fields.FieldInfo):
        return (
            param.annotation,
            param.default
        )
    else:
        return (
            param.annotation,
            pydantic.Field(
                default=param.default if param.default is not inspect.Parameter.empty else ...,
                title=param.name,
                **kwargs,
            ),
        )


def signature_to_model(f: t.Callable,
                       __config__: t.Type[pydantic.BaseConfig] = TeleporterConfig,
                       extra_params: t.Dict[str, type] = None) -> t.Tuple[type, type]:
    """
    Convert the signature of a function to two pydantic models: one for the input and another one for the output.

    Arguments starting with ``_`` are ignored.

    :param f: The function to use the signature of.
    :param __config__: The config the pydantic model should use.
    :param extra_params: Extra parameters to be added to the model.
    :return: A tuple consisting of the input model and the output model.
    """
    if extra_params is None:
        extra_params = {}

    signature: inspect.Signature = inspect.signature(f)

    model_params = {
        key: parameter_to_field(value)
        for key, value in signature.parameters.items()
        if not key.startswith("_")
    }

    input_model = pydantic.create_model(f"{f.__name__}Input",
                                        __config__=TeleporterConfig,
                                        **model_params,
                                        **extra_params)
    output_model = pydantic.create_model(f"{f.__name__}Output",
                                         __config__=TeleporterConfig,
                                         __root__=(signature.return_annotation, pydantic.Field(..., title="Returns")))

    return input_model, output_model


def split_kwargs(**kwargs) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
    """
    Split the kwargs passed to this function in two different :class:`dict`, based on whether their name starts with
    ``_`` or not.

    :return: A tuple of :class:`dict`, where the second contains the ones starting with ``_``, and the first contains
             the rest.
    """
    model_params = {}
    extra_params = {}
    for key, value in kwargs.items():
        if key.startswith("__"):
            raise ValueError("A keyword argument has a key that starts with __.")
        elif key.startswith("_"):
            extra_params[key] = value
        else:
            model_params[key] = value
    return model_params, extra_params


def teleport_in(__model: type, **kwargs):
    """
    Validate the kwargs passed to this function through the passed :mod:`pydantic` ``__model``.

    :param __model: The model that should be used to validate the kwargs.
    :return: The created model.
    :raises .exc.InputValidationError: If the kwargs fail the validation.
    """
    try:
        return __model(**kwargs)
    except pydantic.ValidationError as e:
        raise exc.InTeleporterError(errors=e.raw_errors, model=e.model)


def teleport_out(__model: type, value: Value) -> Value:
    """
    Validate a single value passed to this function through the ``__root__`` of the passed :mod:`pydantic` ``__model``.

    :param __model: The model that should be used to validate the value.
    :param value: The value that should be validated.
    :return: The value passed as result, after being unwrapped from the created model.
    :raises .exc.OutputValidationError: If the value fails the validation.
    """
    try:
        return __model(__root__=value).__root__
    except pydantic.ValidationError as e:
        raise exc.OutTeleporterError(errors=e.raw_errors, model=e.model)


def teleport(__config__: t.Type[pydantic.BaseConfig] = TeleporterConfig,
             is_async: bool = False,
             validate_input: bool = True,
             validate_output: bool = True):
    """
    A factory that returns a decorator which validates a function's passed arguments and its returned value
    using a :mod:`pydantic` model.

    .. warning:: By using this, the function will stop accepting positional arguments, and will only accept
                 keyword arguments.

    :param __config__: The config the pydantic model should use.
    :param is_async: Whether the decorated function is async or not.
    :param validate_input: Whether the input parameters should be validated or not.
    :param validate_output: Whether the return value should be validated or not.
    :return: The decorator to validate the input.

    .. seealso:: :func:`.signature_to_model`
    """
    def decorator(f: t.Callable):
        # noinspection PyPep8Naming
        InputModel, OutputModel = signature_to_model(f, __config__=__config__)

        if is_async:
            @functools.wraps(f)
            async def decorated(**kwargs):
                if validate_input:
                    model_kwargs, extra_kwargs = split_kwargs(**kwargs)
                    model_kwargs = teleport_in(__model=InputModel, **kwargs).dict()
                    kwargs = {**model_kwargs, **extra_kwargs}
                result = await f(**kwargs)
                if validate_output:
                    result = teleport_out(__model=OutputModel, value=result)
                return result
        else:
            @functools.wraps(f)
            def decorated(**kwargs):
                if validate_input:
                    model_kwargs, extra_kwargs = split_kwargs(**kwargs)
                    model_kwargs = teleport_in(__model=InputModel, **kwargs).dict()
                    kwargs = {**model_kwargs, **extra_kwargs}
                result = f(**kwargs)
                if validate_output:
                    result = teleport_out(__model=OutputModel, value=result)
                return result

        return decorated

    return decorator


__all__ = (
    "TeleporterConfig",
    "teleport",
)



from royalnet.royaltyping import *
import abc

__all__ = (
    "Challenge",
    "TrueChallenge",
)


class Challenge(metaclass=abc.ABCMeta):
    """A filter for inputs passed to a :class:`.Campaign`."""

    @abc.abstractmethod
    def filter(self, data: Any) -> bool:
        """Decide if the data should be skipped or not."""
        raise NotImplementedError()


class TrueChallenge(Challenge):
    """A :class:`.Challenge` which always returns :data:`True`."""

    async def filter(self, data: Any) -> bool:
        """Decide if the data should be skipped or not."""
        return True



import asyncio
import pytest


async def sleep_and_raise():
    await asyncio.sleep(0.001)
    raise Exception("Please except this gift!")


@pytest.mark.asyncio
class TestAsyncio:
    async def test_exception(self):
        with pytest.raises(Exception):
            await sleep_and_raise()



"""
This module adds some new common royaltyping to the default typing package.

It should be imported with: ::

    from royalnet.royaltyping import *

"""

from typing import *
# noinspection PyUnresolvedReferences
from typing import IO, TextIO, BinaryIO

if TYPE_CHECKING:
    from royalnet.engineer import Sentry


JSONScalar = Union[
    None,
    float,
    int,
    str,
]
"""
A non-recursive JSON value: either :data:`None`, a :class:`float`, a :class:`int` or a :class:`str`.
"""

JSON = Union[
    JSONScalar,
    List["JSON"],
    Dict[str, "JSON"],
]
"""
A recursive JSON value: either a :data:`.JSONScalar`, or a :class:`list` of :data:`.JSON` objects, or a :class:`dict` 
of :class:`str` to :data:`.JSON` mappings. 
"""

Adventure = Generator[
    Union[
        Tuple[
            Optional["Challenge"],
            ...
        ],
        "Adventure",
        "Challenge",
        None,
    ],
    Any,
    Any,
]
"""
A generator yielding either:

* a :class:`tuple` of a :class:`.Challenge` and extra data;
* a single :class:`.Challenge`;for SQLAlchemy.
* another :data:`.Adventure`;
* :data:`None`. 
"""

AsyncAdventure = AsyncGenerator[
    Union[
        Tuple[
            Optional["AsyncChallenge"],
            ...
        ],
        "AsyncAdventure",
        "AsyncChallenge",
        None,
    ],
    Any,
]
"""
An async generator yielding either:

* a :class:`tuple` of an :class:`.AsyncChallenge` and extra data;
* a single :class:`.AsyncChallenge`;
* another :data:`.AsyncAdventure`;
* :data:`None`. 
"""


class Conversation(Protocol):
    def __call__(self, _sentry: "Sentry", **kwargs) -> Awaitable["Conversation"]:
        ...



from ..exc import RoyalnetException


class CampaignsError(RoyalnetException):
    """
    An error related to the campaigns module.
    """


class ChallengeFailedError(CampaignsError):
    """
    The data passed to the Campaign (or its async equivalent) failed the challenge.
    """


__all__ = (
    "CampaignsError",
    "ChallengeFailedError",
)



from ..exc import RoyalnetException


class ScrollsException(RoyalnetException):
    """An exception raised by the scrolls module."""


class NotFoundError(ScrollsException, KeyError):
    """The requested config key was not found."""


class MissingConfigFileError(ScrollsException, IOError):
    """The config file does not exist, and ``require_file`` is set to :data:`True`."""


class InvalidFormatError(ScrollsException):
    """The requested config key is not valid."""


class ParseError(ScrollsException):
    """The config value could not be parsed correctly."""


class InvalidFileType(ParseError):
    """The type of the specified config file is not currently supported."""


__all__ = (
    "ScrollsException",
    "NotFoundError",
    "MissingConfigFileError",
    "InvalidFormatError",
    "ParseError",
    "InvalidFileType",
)



import pytest
import inspect
import pydantic
import pydantic.fields
import royalnet.engineer.teleporter as tp


@pytest.fixture
def my_function():
    # noinspection PyUnusedLocal
    def f(*, big_f: str, _hidden: int) -> int:
        return _hidden
    return f


def test_parameter_to_field(my_function):
    signature = inspect.signature(my_function)
    parameter = signature.parameters["big_f"]
    t, fieldinfo = tp.parameter_to_field(parameter)
    assert isinstance(fieldinfo, pydantic.fields.FieldInfo)
    assert fieldinfo.default is ...
    assert fieldinfo.title == parameter.name == "big_f"


def test_signature_to_model(my_function):
    # noinspection PyPep8Naming
    InputModel, OutputModel = tp.signature_to_model(my_function)
    assert callable(InputModel)

    model = InputModel(big_f="banana")
    assert isinstance(model, pydantic.BaseModel)
    assert model.big_f == "banana"
    assert model.dict() == {"big_f": "banana"}

    with pytest.raises(pydantic.ValidationError):
        InputModel()

    model = InputModel(big_f="exists", _hidden="no")
    assert isinstance(model, pydantic.BaseModel)
    assert model.big_f == "exists"
    with pytest.raises(AttributeError):
        _ = model._hidden

    with pytest.raises(pydantic.ValidationError):
        InputModel(big_f=...)

    model = OutputModel(__root__=42)
    assert isinstance(model, pydantic.BaseModel)
    assert model.__root__ == 42
    assert model.dict() == {"__root__": 42}

    with pytest.raises(pydantic.ValidationError):
        OutputModel()

    with pytest.raises(pydantic.ValidationError):
        OutputModel(__root__="sì")


# noinspection PyTypeChecker
class TestTeleporter:
    def test_standard_function(self):
        @tp.teleport()
        def standard_function(a: int, b: int, _return_str: bool = False) -> int:
            if _return_str:
                return "You asked me this."
            return a+b

        assert standard_function(a=1, b=2) == 3
        assert standard_function(a="1", b="2") == 3
        assert standard_function(a="1", b=2) == 3

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a="sì", b="no")

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a=1, b="no")

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a="sì", b=2)

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a="sì")

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(b="sì")

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a=1, b=2, _return_str=True)

        with pytest.raises(TypeError):
            _ = standard_function(1, 2)

    @pytest.mark.asyncio
    async def test_async_function(self):
        @tp.teleport(is_async=True)
        async def async_function(a: int, b: int, _return_str: bool = False) -> int:
            if _return_str:
                return "You asked me this."
            return a+b

        assert await async_function(a=1, b=2) == 3
        assert await async_function(a="1", b="2") == 3
        assert await async_function(a="1", b=2) == 3

        with pytest.raises(pydantic.ValidationError):
            _ = await async_function(a="sì", b="no")

        with pytest.raises(pydantic.ValidationError):
            _ = await async_function(a=1, b="no")

        with pytest.raises(pydantic.ValidationError):
            _ = await async_function(a="sì", b=2)

        with pytest.raises(pydantic.ValidationError):
            _ = await async_function(a="sì")

        with pytest.raises(pydantic.ValidationError):
            _ = await async_function(b="sì")

        with pytest.raises(pydantic.ValidationError):
            _ = await async_function(a=1, b=2, _return_str=True)

        with pytest.raises(TypeError):
            _ = await async_function(1, 2)

    def test_only_input(self):
        @tp.teleport(validate_output=False)
        def standard_function(a: int, b: int, _return_str: bool = False) -> int:
            if _return_str:
                return "You asked me this."
            return a+b

        assert standard_function(a=1, b=2) == 3
        assert standard_function(a="1", b="2") == 3
        assert standard_function(a="1", b=2) == 3

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a="sì", b="no")

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a=1, b="no")

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a="sì", b=2)

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a="sì")

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(b="sì")

        assert standard_function(a=1, b=2, _return_str=True) == "You asked me this."

        with pytest.raises(TypeError):
            _ = standard_function(1, 2)

    def test_only_output(self):
        @tp.teleport(validate_input=False)
        def standard_function(a: int, b: int, _return_str: bool = False) -> int:
            if _return_str:
                return "You asked me this."
            return a+b

        assert standard_function(a=1, b=2) == 3
        assert standard_function(a="1", b="2") == 12

        with pytest.raises(TypeError):
            assert standard_function(a="1", b=2) == 3

        with pytest.raises(pydantic.ValidationError):
            assert standard_function(a="sì", b="no") == "sìno"

        with pytest.raises(TypeError):
            _ = standard_function(a=1, b="no")

        with pytest.raises(TypeError):
            _ = standard_function(a="sì", b=2)

        with pytest.raises(TypeError):
            _ = standard_function(a="sì")

        with pytest.raises(TypeError):
            _ = standard_function(b="sì")

        with pytest.raises(pydantic.ValidationError):
            _ = standard_function(a=1, b=2, _return_str=True)

        with pytest.raises(TypeError):
            _ = standard_function(1, 2)



"""

"""

from __future__ import annotations
import royalnet.royaltyping as t

import abc
import datetime
import sqlalchemy.orm

from . import exc


class Bullet(metaclass=abc.ABCMeta):
    """
    The abstract base class for Bullet data models.
    """

    @abc.abstractmethod
    def __hash__(self) -> int:
        """
        :return: A value that uniquely identifies the object in this Python interpreter process.
        """
        raise NotImplementedError()


class Message(Bullet, metaclass=abc.ABCMeta):
    """
    An abstract class representing a chat message.
    """

    async def text(self) -> t.Optional[str]:
        """
        :return: The raw text contents of the message.
        :raises .exc.NotSupportedError: If the frontend does not support text messages.
        """
        raise exc.NotSupportedError()

    async def timestamp(self) -> t.Optional[datetime.datetime]:
        """
        :return: The :class:`datetime.datetime` at which the message was sent.
        :raises .exc.NotSupportedError: If the frontend does not support timestamps.
        """
        raise exc.NotSupportedError()

    async def reply_to(self) -> t.Optional[Message]:
        """
        :return: The :class:`.Message` this message is a reply to.
        :raises .exc.NotSupportedError: If the frontend does not support replies.
        """
        raise exc.NotSupportedError()

    async def channel(self) -> t.Optional[Channel]:
        """
        :return: The :class:`.Channel` this message was sent in.
        :raises .exc.NotSupportedError: If the frontend does not support channels.
        """
        raise exc.NotSupportedError()


class Channel(Bullet, metaclass=abc.ABCMeta):
    """
    An abstract class representing a channel where messages can be sent.
    """

    async def name(self) -> t.Optional[str]:
        """
        :return: The name of the message channel, such as the chat title.
        :raises .exc.NotSupportedError: If the frontend does not support channel names.
        """
        raise exc.NotSupportedError()

    async def topic(self) -> t.Optional[str]:
        """
        :return: The topic (description) of the message channel.
        :raises .exc.NotSupportedError: If the frontend does not support channel topics / descriptions.
        """
        raise exc.NotSupportedError()

    async def users(self) -> t.List[User]:
        """
        :return: A :class:`list` of :class:`.User` who can read messages sent in the channel.
        :raises .exc.NotSupportedError: If the frontend does not support such a feature.
        """
        raise exc.NotSupportedError()


class User(Bullet, metaclass=abc.ABCMeta):
    """
    An abstract class representing a user who can read or send messages in the chat.
    """

    async def name(self) -> t.Optional[str]:
        """
        :return: The user's name.
        :raises .exc.NotSupportedError: If the frontend does not support usernames.
        """
        raise exc.NotSupportedError()

    async def database(self, session: sqlalchemy.orm.Session) -> t.Any:
        """
        :param session: A :class:`sqlalchemy.orm.Session` instance to use to fetch the database entry.
        :return: The database entry for this user.
        """
        raise exc.NotSupportedError()


__all__ = (
    "Bullet",
    "Message",
    "Channel",
    "User",
)



import sqlalchemy


def ieq(one, two):
    """
    Create a case-insensitive equality filter to be used in :meth:`sqlalchemy.orm.query.Query.filter`.

    Equal to: ::

        lower(one) == lower(two)

    """
    return sqlalchemy.func.lower(one) == sqlalchemy.func.lower(two)


def ineq(one, two):
    """
    Create a case-insensitive inequality filter to be used in :meth:`sqlalchemy.orm.query.Query.filter`.

    Equal to: ::

        lower(one) != lower(two)

    """
    return sqlalchemy.func.lower(one) != sqlalchemy.func.lower(two)


__all__ = (
    "ieq",
    "ineq",
)



class ColRepr:
    """
    A mixin that can be added to a declared class to display all columns of the table with their values in the
    ``__repr__``.
    """

    def __repr__(self):
        columns = [c.name for c in self.__class__.__table__.columns]
        args = [f"{c}={repr(self.__getattribute__(c))}" for c in columns]
        return f"{self.__class__.__qualname__}({', '.join(args)})"


__all__ = (
    "ColRepr",
)
